;;; $Id: aop-3-7.scm,v 1.6 2006/04/25 17:33:02 leavens Exp $
;; Aspect-oriented language interpreter (based on ch3-7.scm).

(module aop-3-7 (lib "typedscm.ss" "typedscm")
  
(provide read-eval-print eval-program run
         scan&parse just-scan
         run-all run-one equal-external-reps?)
  
(require (lib "environment-3-7.scm" "lib342")
         (lib "reference-3-7.scm" "lib342")
         (lib "slib-hashtab.scm" "typedscm")
         (lib "test-suite.scm" "lib342")
         (lib "relax-type-checking.scm" "typedscm"))

;;;;;;;;;;;;;;;; top level and tests ;;;;;;;;;;;;;;;;

(deftype run (-> (string) datum))
(define run
  (lambda (string)
    (eval-program (scan&parse string))))

;; Use aop-3-7.tst instead
(deftype run-all (-> () void))
(define run-all
  (lambda ()
    (run-experiment run use-execution-outcome
      '(aop-3-7) (all-tests))))

(deftype run-one (-> (symbol) datum))
(define run-one
  (lambda (test-name)
    (run-test run test-name)))

;; needed for testing
(deftype equal-external-reps? (-> (datum datum) boolean))
(define equal-external-reps?
  (lambda (x y)
    (equal? x y)))

;;;;;;;;;;;;;;;; grammatical specification ;;;;;;;;;;;;;;;;

(define the-lexical-spec
  '((whitespace (whitespace) skip)
    (comment ("%" (arbno (not #\newline))) skip)
    (identifier
      (letter (arbno (or letter digit "_" "-" "?")))
      symbol)
    (number (digit (arbno digit)) number)))

(define the-grammar
  '((program ((arbno advice) "run" expression) a-program)

    ;; added for AOP
    (advice ("around" pcd "do" expression) around-adv)
    ; (advice ("before" pcd "do" expression) before-adv)
    ; (advice ("after" pcd "do" expression) after-adv)

    (pcd ("get" idpattern) get-pcd)
    (pcd ("set" idpattern) set-pcd)

    (idpattern (identifier) id-pat)
    (idpattern ("*") wild-pat)
    
    (expression (number) lit-exp)
    (expression (identifier) var-exp)   
    (expression
      (primitive "(" (separated-list expression ",") ")")
      primapp-exp)
    (expression
      ("if" expression "then" expression "else" expression)
      if-exp)
    (expression
      ("let" (arbno  identifier "=" expression) "in" expression)
      let-exp)
    (expression
      ("proc" "(" (separated-list identifier ",") ")" expression)
      proc-exp)
    (expression
      ("(" expression (arbno expression) ")")
      app-exp)
    (expression
      ("begin" expression (arbno ";" expression) "end")
      begin-exp)
    (expression                         
      ("letrec"
        (arbno identifier "(" (separated-list identifier ",") ")"
          "=" expression)
        "in" expression)
      letrec-exp)
    (expression ("set" identifier "=" expression) varassign-exp)
    ;; added for AOP
    (expression ("proceed") proceed-exp)
    (expression ("LHS" "=" expression) varassign-LHS-exp)
    (expression ("RHS") RHS-exp)

    (primitive ("+")     add-prim)
    (primitive ("-")     subtract-prim)
    (primitive ("*")     mult-prim)
    (primitive ("add1")  incr-prim)
    (primitive ("sub1")  decr-prim)
    (primitive ("zero?") zero-test-prim)
    ))

(deftype show-the-datatypes (-> () (list-of datum)))
(define show-the-datatypes
  (lambda ()
    (sllgen:make-define-datatypes the-lexical-spec the-grammar)
    (sllgen:list-define-datatypes the-lexical-spec the-grammar)))

;; The following is generated by calling show-the-datatypes
;; (and then removing the outer lists, indenting, and renaming fields).
(define-datatype
  program
  program?
  (a-program
   (advice-defs (list-of advice?)) ; added for AOP
   (exp expression?)))

;; added for AOP
(define-datatype advice advice?
  (around-adv (pcd pcd?)
              (body expression?))
;   (before-adv (pcd pcd?)
;             (body expression?))
;   (after-adv (pcd pcd?)
;            (body expression?))
  )

;; added for AOP
(define-datatype pcd pcd?
  (get-pcd (id-pat idpattern?))
  (set-pcd (id-pat idpattern?)))

;; added for AOP
(define-datatype idpattern idpattern?
  (id-pat (id symbol?))
  (wild-pat))

(define-datatype
  expression
  expression?
  (lit-exp (datum number?))
  (var-exp (id symbol?))
  (primapp-exp
   (prim primitive?)
   (rands (list-of expression?)))
  (if-exp
   (test-exp expression?)
   (true-exp expression?)
   (false-exp expression?))
  (let-exp
   (ids (list-of symbol?))
   (rands (list-of expression?))
   (body expression?))
  (proc-exp
   (ids (list-of symbol?))
   (body expression?))
  (app-exp
   (rator expression?)
   (rands (list-of expression?)))
  (begin-exp
   (first expression?)
   (rest (list-of expression?)))
  (letrec-exp
   (proc-names (list-of symbol?))
   (idss (list-of (list-of symbol?)))
   (bodies (list-of expression?))
   (letrec-body expression?))
  (varassign-exp
   (id symbol?)
   (rhs-exp expression?))
  ;; added for AOP, visible to programmers
  (proceed-exp)
  (varassign-LHS-exp
   (rhs-exp expression?))
  (RHS-exp)
  ;; The following are not usable by programmers
  ;; (they aren't in the concrete syntax),
  ;; but are used in the interprer for aspect weaving
  (unadvisable-var-exp (id symbol?))
  (unadvisable-varassign-exp
   (id symbol?)
   (rhs-exp expression?))
  (continue-weave-get-exp
   (remaining-advice (list-of expression?))
   (base-exp expression?))
  (continue-weave-set-exp
   (remaining-advice (list-of expression?))
   (base-exp expression?)
   (id symbol?)
   (rhs-exp expression?))
  )

(define-datatype
  primitive
  primitive?
  (add-prim)
  (subtract-prim)
  (mult-prim)
  (incr-prim)
  (decr-prim)
  (zero-test-prim)
  )

(deftype scan&parse (-> (string) program))
(define scan&parse
  (lambda (ptext)
    ((sllgen:make-string-parser the-lexical-spec the-grammar)
     ptext)))

(deftype just-scan (forall (TOK) (-> (string) (list-of TOK))))
(define just-scan
  (lambda (ptext)
    ((sllgen:make-string-scanner the-lexical-spec the-grammar)
     ptext)))

(deftype read-eval-print (-> () poof))
(define read-eval-print
  (lambda ()
    ((sllgen:make-rep-loop
      "--> "
      (lambda (pgm) (eval-program pgm))
      (sllgen:make-stream-parser 
       the-lexical-spec
       the-grammar)))))

;;; Domains for this interpreter:
;;;
;;; Expressed-Value = Number + ProcVal + List(Expressed-Value)
;;; Denoted-Value = Ref(Expressed-Value)

;;; ADTs used

;;;;;;;;;;;;;;;; ProcVal ;;;;;;;;;;;;;;;;

(deftype procval? (type-predicate-for procval))
(deftype closure (-> ((list-of symbol) expression environment) procval))
(deftype apply-procval
  (-> (procval (list-of Expressed-Value)) Expressed-Value))

(define-datatype procval procval?
  (closure 
    (ids (list-of symbol?)) 
    (body expression?)
    (env environment?)))

(define apply-procval
  (lambda (proc args)
    (cases procval proc
      (closure (ids body env)
        (eval-expression body (extend-env ids args env))))))

;;;;;;;;;;; advice tables ;;;;;;;;;;;;;;;;;;;

(deftype empty-advice-table (-> () advice-table))
(deftype add-advice! (-> (advice-table advice) void))
(deftype around-advice-for-get (-> (advice-table symbol) (list-of expression)))
(deftype around-advice-for-set (-> (advice-table symbol) (list-of expression)))

(define hash-to-expressions?
  (has-type-trusted
   (type-predicate-for (slib:hash-table symbol (list-of expression)))
   (lambda (dat)
     (slib:hash-table? dat))))
  
(define-datatype advice-table advice-table?
  (an-advice-table
   (get-patterns (vector-of (list-of expression?)))
   (get-ids hash-to-expressions?)
   (set-patterns (vector-of (list-of expression?)))
   (set-ids hash-to-expressions?)))

(define empty-advice-table
  (lambda ()
    (an-advice-table (vector '()) (make-hash-table 7)
                     (vector '()) (make-hash-table 7))))

(define primitive-hash-lookup
  (lambda (hashtab key)
    ((hash-inquirer eq?) hashtab key)))

(define hash-lookup
  (lambda (htab id)
    (let ((res (primitive-hash-lookup htab id)))
      (if (has-type-trusted boolean res)
          res
          '()))))

(define hash-add!
  (lambda (hashtab key val)
    ((hash-associator eq?) hashtab key val)))

(define add-advice!
  (lambda (adv-tab adv)
    (cases advice-table adv-tab
       (an-advice-table (get-patterns get-ids set-patterns set-ids)
         (cases advice adv
           (around-adv (p body)
             (cases pcd p
               (get-pcd (idp)
                 (add-to-tables! get-patterns get-ids idp body))
               (set-pcd (idp)
                 (add-to-tables! set-patterns set-ids idp body))
           (else (error "add-advice! case not implemented")))))))))
  
(deftype add-to-tables! (-> ((vector-of (list-of expression))
                             (slib:hash-table symbol (list-of expression))
                             idpattern
                             expression)
                            void))
(define add-to-tables!
  (lambda (patterns ids idp body)
    (cases idpattern idp
       (id-pat (id) 
          (let ((adv-for-id (test-type (list-of expression?) (hash-lookup ids id))))
            (ignore (hash-add! ids id
                       (append adv-for-id (list body))))))
       (else (vector-set! patterns 0
                          (append (vector-ref patterns 0) (list body)))))))

(define around-advice-for-get
  (lambda (adv-tab id)
    (cases advice-table adv-tab
       (an-advice-table (get-patterns get-ids set-patterns set-ids)
           (append (vector-ref get-patterns 0)
                   (test-type (list-of expression?) (hash-lookup get-ids id)))))))

(define around-advice-for-set
  (lambda (adv-tab id)
    (cases advice-table adv-tab
       (an-advice-table (get-patterns get-ids set-patterns set-ids)
           (append (vector-ref set-patterns 0)
                   (test-type (list-of expression?) (hash-lookup set-ids id)))))))

;;; Anonymous figure : page 80

(deftype true-value? (-> (Expressed-Value) boolean))
(define true-value?
  (lambda (x)
    (not (zero? (expressed->number x)))))

;;;;;;;;; Expressed-Value ;;;;;;;;;;;;;;;;;;;;;;;;;

;; upcasts
(deftype number->expressed (-> (number) Expressed-Value))
(deftype procval->expressed (-> (Procval) Expressed-Value))
(deftype list->expressed (-> ((list-of Expressed-Value)) Expressed-Value))

;; downcasts
(deftype expressed->number (-> (Expressed-Value) number))
(deftype expressed->procval (-> (Expressed-Value) Procval))
(deftype expressed->list (-> (Expressed-Value) (list-of Expressed-Value)))

;; debugging
(deftype expressed->printable (-> (Expressed-Value) datum))

;; tests
(deftype number->expressed? (-> (Expressed-Value) boolean))
(deftype procval->expressed? (-> (Expressed-Value) boolean))
(deftype list->expressed? (-> (Expressed-Value) boolean))

(define-datatype Expressed-Value expval?
  (number->expressed (num number?))
  (procval->expressed (pv procval?))
  (list->expressed (lst (list-of expval?))))

(define expressed->number
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) num)
      (else (error "expressed->number passed non-number argument: " ev)))))

(define expressed->procval
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) pv)
      (else (error "expressed->procval passed non-procval argument: " ev)))))

(define expressed->list
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) lst)
      (else (error "expressed->list passed non-list argument: " ev)))))

(define expressed->printable
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) (has-type datum num))
      (procval->expressed (pv)
        (cases procval pv
          (closure (ids body env)
           (has-type datum (list 'closure ids body "<env>")))))
      (list->expressed (lst)
         (has-type datum (map expressed->printable lst))))))

(define number->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) #t)
      (else #f))))

(define procval->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) #t)
      (else #f))))

(define list->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) #t)
      (else #f))))

;; adaptation of environment-3-7 to this file
(deftype extend-env-recursively
  (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
       environment)
      environment))
(define extend-env-recursively
  (extend-env-recursively-maker
   (lambda (ids body env)
     (procval->expressed (closure ids body env)))
   (procval->expressed (closure '() (lit-exp 0) (empty-env)))))

;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

;; Added to track advice
(deftype *advice-table* advice-table)
(define *advice-table* (empty-advice-table))

(deftype eval-program (-> (program) datum))
(define eval-program 
  (lambda (pgm)
    ;; prevent different program runs from interfering with each other
    (set! *advice-table* (empty-advice-table))
    (cases program pgm
      (a-program (advice-defs body)
        (begin
          (for-each (lambda (adv) (add-advice! *advice-table* adv))
                    advice-defs)
          (expressed->printable (eval-expression body (init-env))))))))

(deftype dynamically-weave-around-get
  (-> ((list-of expression) expression environment) Expressed-Value))
(define dynamically-weave-around-get
  (lambda (advice-exps base-exp env)
    (if (null? advice-exps)
        (eval-expression base-exp env)
        (eval-expression
         (car advice-exps)
         (extend-env
          '(proceed)
          (list (procval->expressed
                 (closure '()
                          (continue-weave-get-exp
                           (cdr advice-exps)
                           base-exp)
                          env)))
          env)))))
                                      
(deftype dynamically-weave-around-set
  (-> ((list-of expression) expression symbol expression environment)
      Expressed-Value))
(define dynamically-weave-around-set
  (lambda (advice-exps base-exp id rhs-exp env)
    (if (null? advice-exps)
        (eval-expression base-exp env)
        (eval-expression
         (car advice-exps)
         (extend-env
          '(proceed LHS RHS)
          (list
           ;; Binding for proceed below.  When set advice proceeds,
           ;; the interpreter find the closure bound to proceed, and
           ;; runs it, which runs the next advice or the base expression.
           (procval->expressed
                 (closure '()
                          (continue-weave-set-exp
                           (cdr advice-exps) base-exp
                           id rhs-exp)
                          env))
           ;; Binding for LHS below.  When an expression of the form
           ;;    set LHS = <expression>
           ;; is encountered, we find the closure below, and call it,
           ;; passing it the value of the given <expression>.
           ;; The closure below will assign that value to the saved LHS name.
           (procval->expressed
            (closure (list '$rhs)
                     (unadvisable-varassign-exp id (unadvisable-var-exp '$rhs))
                     env))
           ;; Binding for RHS below.  When an expression of the form
           ;;    RHS
           ;; is encountered, we find the closure below, and call it
           ;; to obtain the value of the saved rhs-exp.
           (procval->expressed
            (closure '()
                     rhs-exp
                     env)))
          env)))))

(deftype eval-expression (-> (expression environment) Expressed-Value))
(define eval-expression 
  (lambda (exp env)
    (cases expression exp
      (lit-exp (datum)
        (number->expressed datum))
      (var-exp (id)
        (let ((applicable-around-advice
               (around-advice-for-get *advice-table* id)))
          (dynamically-weave-around-get applicable-around-advice
                                    (unadvisable-var-exp id)
                                    env)))
      (unadvisable-var-exp (id)
        (apply-env env id))
      (proceed-exp ()
        (if (defined-in-env? env 'proceed)
            (let ((proceed-closure
                   (expressed->procval (apply-env env 'proceed))))
              (apply-procval proceed-closure '()))
            (eopl:error 'eval-expression
                        "proceed may only be used from within around advice")))
      (continue-weave-get-exp (remaining-advice base-exp)
        (dynamically-weave-around-get remaining-advice base-exp env))
      (continue-weave-set-exp (remaining-advice base-exp id rhs-exp)
        (dynamically-weave-around-set remaining-advice base-exp id rhs-exp env))
      (primapp-exp (prim rands)
        (let ((args (eval-rands rands env)))
          (apply-primitive prim args)))
      (if-exp (test-exp true-exp false-exp)
        (if (true-value? (eval-expression test-exp env))
          (eval-expression true-exp env)
          (eval-expression false-exp env)))
      (begin-exp (exp1 exps)
        (let loop ((acc (eval-expression exp1 env))
                   (exps exps))
          (if (null? exps) acc
            (loop (eval-expression (car exps) env) (cdr exps)))))
      (let-exp (ids rands body)
        (let ((args (eval-rands rands env)))
          (eval-expression
           body
           (extend-env ids args env))))
      (proc-exp (ids body) (procval->expressed (closure ids body env)))
      (app-exp (rator rands)
        (let ((proc (eval-expression rator env))
              (args (eval-rands rands env)))
          (if (procval->expressed? proc)
            (apply-procval (expressed->procval proc) args)
            (eopl:error 'eval-expression
              "Attempt to apply non-procedure ~s" proc))))
      (letrec-exp (proc-names idss bodies letrec-body)
        (eval-expression letrec-body
          (extend-env-recursively proc-names idss bodies env)))     
      (varassign-exp (id rhs-exp)
        (let ((applicable-around-advice
               (around-advice-for-set *advice-table* id)))
          (dynamically-weave-around-set
           applicable-around-advice
           (unadvisable-varassign-exp id rhs-exp)
           id rhs-exp env)))
      (unadvisable-varassign-exp (id rhs-exp)
        (begin
          (setref! 
            (apply-env-ref env id)
            (eval-expression rhs-exp env))
          (number->expressed 1)))
      (varassign-LHS-exp (rhs-exp)
        (if (defined-in-env? env 'LHS)
            (let ((action-closure (apply-env env 'LHS))
                  (rhs-val (eval-expression rhs-exp env)))
              (apply-procval (expressed->procval action-closure)
                             (list rhs-val)))
            (eopl:error
             'eval-expression
             "LHS = <expression> can only be used in advice on set")))
      (RHS-exp ()
        (if (defined-in-env? env 'RHS)
            (apply-procval (expressed->procval (apply-env env 'RHS)) '())
            (eopl:error
             'eval-expression
             "RHS can only be used in advice on set")))
      (else (eopl:error 'eval-expression "Missing case:~s" exp))
      )))

(deftype eval-rands
  (-> ((list-of expression) environment) (list-of Expressed-Value)))
(define eval-rands
  (lambda (rands env)
    (map (lambda (x) (eval-rand x env)) rands)))

(deftype eval-rand (-> (expression environment) Expressed-Value))
(define eval-rand
  (lambda (rand env)
    (eval-expression rand env)))

(deftype apply-primitive
  (-> (primitive (list-of Expressed-Value)) Expressed-Value))
(define apply-primitive
  (lambda (prim args)
    (cases primitive prim
      (add-prim  () (number->expressed
                     (+ (expressed->number (car args))
                        (expressed->number (cadr args)))))
      (subtract-prim () (number->expressed
                         (- (expressed->number (car args))
                            (expressed->number (cadr args)))))
      (mult-prim  () (number->expressed
                      (* (expressed->number (car args))
                         (expressed->number (cadr args)))))
      (incr-prim  () (number->expressed
                      (+ (expressed->number (car args)) 1)))
      (decr-prim  () (number->expressed
                      (- (expressed->number (car args)) 1)))
      (zero-test-prim () (number->expressed
                          (if (zero? (expressed->number (car args))) 1 0)))
      )))

(deftype init-env (-> () environment))
(define init-env 
  (lambda ()
    (extend-env
      '(i v x)
      (map number->expressed '(1 5 10))
      (empty-env))))

) ;; end module
