;;; $Id: tc-compute-cxr-types.scm,v 1.10 2006/01/18 23:17:27 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Generate types for built-in Scheme procedures cxr (e.g., car, cdr, cadr)
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.
;;;          Adapted by Brian Dorn to work with new representation of typing
;;;          and unification

(module tc-compute-cxr-types (lib "typedscm.ss" "typedscm")

  (provide *tc:cxr-output-type-env* tc:print-cxr-output-type-env)
  
  (require (lib "tc-global-var-output-types.scm" "typedscm")
	   (lib "tc-scheme-unparser.scm" "typedscm")
	   (lib "tc-scheme-parser.scm" "typedscm")
	   (lib "tc-types.scm" "typedscm")
	   (lib "tc-type-translate.scm" "typedscm")
           (lib "tc-scheme-annotate.scm" "typedscm"))

  (deftype tc:compose-cxr-types 
    (-> (tc:type-expr tc:type-expr) tc:type-expr))
  (define tc:compose-cxr-types
    (lambda (left right)
      ;; REQUIRES: left and right are "and types"
      (let ((left-types (tc:intersection-type-expr->conjoined-types left))
	    (right-types (tc:intersection-type-expr->conjoined-types right)))
	(letrec ((append-map
		  (lambda (f ls)
		    (if (null? ls)
			'()
			(append (f (car ls))
				(append-map f (cdr ls)))))))
	  (tc:make-intersection-type-expr
	   (append-map
	    (lambda (left-typ)
	      (append-map
	       (lambda (right-typ)
		 (tc:compose-arrow-types left-typ right-typ))
	       right-types))
	    left-types))))))

  (deftype tc:compose-arrow-types
    (-> (tc:type-expr tc:type-expr) (list tc:type-expr)))
  (define tc:compose-arrow-types
    (lambda (left-type right-type)
      ;; REQUIRES: both left-type and right-type are one-argument function types
      ;; ENSURES: result is a list containing a single
      ;;          function-type-expr, if the two types can be composed,
      ;;          otherwise an empty list.
      (let* ((left-uniq-type
	      (tc:type-expr-uniquify left-type))
	     (right-uniq-type
	      (tc:type-expr-uniquify right-type))
	     ;;(dummy (begin (displayln "left-uniq-type = " left-uniq-type)
	     ;;	           (displayln "right-uniq-type = " right-uniq-type)))
	     (left-result-type
	      (tc:function-type-expr->result-type left-uniq-type))
	     (right-result-type
	      (tc:function-type-expr->result-type right-uniq-type))
	     (right-arg-type
	      (car (tc:function-type-expr->arg-types right-uniq-type)))
	     (left-arg-type
	      (car (tc:function-type-expr->arg-types left-uniq-type))))
	(let ((e ((tc:mgu (tc:type-expr-as-unifiable)) right-result-type left-arg-type)))
	
	  ;; (displayln "tc:compose-arrow-types got env " e)
	  (if (null? e)	;; Test for unification error
	      '()
	      (let ((app (tc:app (car e))))
		(list (tc:type-expr-uniquify
		       ;; use both tc:type-unparse and
		       ;; tc:type-parse below to make sure that all
		       ;; variable names are distinct (not just the
		       ;; numbers) before we use uniquify otherwise 
		       ;; variables with the same name but different
		       ;; numbers are collapsed...
		       (tc:type-parse
			(tc:type-unparse
			 ;;(tc:subst-bindings (something->value e)
			 (tc:make-function-type-expr
			  (list (app right-arg-type))
			  (app left-result-type))))))))))))

  (deftype *tc:cxr-list* (list-of (pair-of (symbol (list-of symbol)))))
  (define *tc:cxr-list*
    '((caar car car)
      (cadr car cdr)
      (cdar cdr car)
      (cddr cdr cdr)

      (caaar car car car)
      (caadr car car cdr)
      (cadar car cdr car)
      (caddr car cdr cdr)
      (cdaar cdr car car)
      (cdadr cdr car cdr)
      (cddar cdr cdr car)
      (cdddr cdr cdr cdr)

      (caaaar car car car car)
      (caaadr car car car cdr)
      (caadar car car cdr car)
      (caaddr car car cdr cdr)
      (cadaar car cdr car car)
      (cadadr car cdr car cdr)
      (caddar car cdr cdr car)
      (cadddr car cdr cdr cdr)
      (cdaaar cdr car car car)
      (cdaadr cdr car car cdr)
      (cdadar cdr car cdr car)
      (cdaddr cdr car cdr cdr)
      (cddaar cdr cdr car car)
      (cddadr cdr cdr car cdr)
      (cdddar cdr cdr cdr car)
      (cddddr cdr cdr cdr cdr)
      ))

  (deftype *tc:cxr-output-type-env* (tc:env-of symbol tc:output-type-expr))
  (define *tc:cxr-output-type-env*
    (letrec ((car-type
	      (has-type
		tc:type-expr
		(tc:type-parse
		 (tc:env-value 'car *tc:pre-global-var-output-type-env*))))
	     (cdr-type
	      (has-type
		tc:type-expr
		(tc:type-parse
		 (tc:env-value 'cdr *tc:pre-global-var-output-type-env*))))
	     (name->type-expr
	      (has-type
		(-> (symbol) tc:type-expr)
		(lambda (name)
		  (if (eq? name 'car) car-type cdr-type))))
	     (names->type-expr
	      (has-type
		(-> ((list symbol)) tc:type-expr)
		(lambda (names-list)
		  (if (null? (cdr names-list))
		      (name->type-expr (car names-list))
		      (tc:compose-cxr-types
		       (name->type-expr (car names-list))
		       (names->type-expr (cdr names-list))))))))
      (tc:env-map
       (lambda (name recipe)
	 (tc:type-unparse (names->type-expr recipe)))
       (tc:assoc-list->env *tc:cxr-list*))))


  (deftype tc:print-cxr-output-type-env (-> () void))
  (define tc:print-cxr-output-type-env
    (lambda ()
      (pretty-print *tc:cxr-output-type-env*)))
  )
