;;; $Id: tc-scheme-token.scm,v 1.6 2006/02/06 01:09:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

(module tc-scheme-token (lib "typedscm.ss" "typedscm")

(provide tc:scheme-token tc:scheme-token? tc:scheme-token-record
	 tc:scheme-token->data tc:scheme-token->position
	 tc:scheme-token->class tc:token-class->string
	 ;; all-from tc-position
	 tc:position tc:position? tc:make-token-position tc:make-tree-position
	 tc:position->start-line tc:position->end-line tc:position->filename
	 tc:position->error-string tc:position->error-lines-string
	 tc:current-scanner-position tc:current-line tc:new-line!
	 tc:current-input-filename
	 tc:start-input-filename! tc:end-input-filename!)

(require (lib "tc-position.scm" "typedscm"))
  
(define-datatype tc:scheme-token tc:scheme-token?
  (tc:scheme-token-record
   (position tc:position?)
   (class symbol?)
   (data datum?)))

(deftype tc:scheme-token->data (-> (tc:scheme-token) datum))
(define tc:scheme-token->data
  (lambda (tok)
    (cases tc:scheme-token tok
	   (tc:scheme-token-record (position class data) data))))

(deftype tc:scheme-token->position (-> (tc:scheme-token) tc:position))
(define tc:scheme-token->position
  (lambda (tok)
    (cases tc:scheme-token tok
	   (tc:scheme-token-record (position class data) position))))

(deftype tc:scheme-token->class (-> (tc:scheme-token) symbol))
(define tc:scheme-token->class
  (lambda (tok)
    (cases tc:scheme-token tok
	   (tc:scheme-token-record (position class data) class))))

(deftype tc:token-class->string (-> (symbol) string))
(define tc:token-class->string
  (lambda (token-class)
    (case token-class
      ((lparen) "(")
      ((rparen) ")")
      ((dot) ".")
      ((end-marker) "<end of file>")
      ((quote-char) "'")
      ((quasiquote-char) "`")
      ((unquote-char) ",")
      ((unquote-splicing-char) ",@")
      ((sharp-lparen) "#(")
      ((number boolean string char symbol)
       (string-append "<" (symbol->string token-class) ">"))
      (else (symbol->string token-class)))))
)
