;;; $Id: tc-token-seq-as-stream.scm,v 1.5 2006/02/06 01:09:10 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Token sequences, represented by streams

(module tc-token-seq-as-stream (lib "typedscm.ss" "typedscm")

(provide tc:token-seq? tc:the-null-token-seq tc:token-seq-null?
         tc:make-token-seq tc:make-token-seq-with-thunk
         tc:token-seq-head tc:token-seq-tail tc:token-seq-for-each
         ;; all-from tc-parsing-stream.scm
         tc:stream-of tc:stream-car tc:stream-cdr tc:make-stream
         tc:the-null-stream tc:stream-null? tc:stream-for-each
         tc:make-input-stream)

(deftype tc:token-seq? (type-predicate-for tc:token-seq))
(deftype tc:the-null-token-seq (-> () tc:token-seq))
(deftype tc:token-seq-null? (-> (tc:token-seq) boolean))
(deftype tc:make-token-seq (-> (tc:scheme-token tc:token-seq) tc:token-seq))
(deftype tc:make-token-seq-with-thunk
  (-> (tc:scheme-token (-> () tc:token-seq))
      tc:token-seq))
(deftype tc:token-seq-head (-> (tc:token-seq) tc:scheme-token))
(deftype tc:token-seq-tail (-> (tc:token-seq) tc:token-seq))
(deftype tc:token-seq-for-each
  (forall (T) (-> ((-> (T) void) tc:token-seq) void)))

(require (lib "tc-scheme-token.scm" "typedscm")
	 (lib "tc-parsing-stream.scm" "typedscm"))

(defrep (tc:token-seq (tc:stream-of tc:scheme-token)))

(define tc:token-seq?
  (lambda (obj)
    ((tc:stream-of tc:scheme-token?) obj)))

(define tc:the-null-token-seq
  (lambda ()
    (tc:the-null-stream)))

(define tc:token-seq-null?
  (lambda (ts)
    (tc:stream-null? ts)))

(define tc:make-token-seq
  (lambda (token seq)
    (tc:make-stream token (lambda () seq))))

(define tc:make-token-seq-with-thunk
  (lambda (t th)
    (tc:make-stream t th)))

(define tc:token-seq-head
  (lambda (ts)
    (tc:stream-car ts)))

(define tc:token-seq-tail
  (lambda (ts)
    (tc:stream-cdr ts)))

(define tc:token-seq-for-each
  (lambda (proc ts)
    (tc:stream-for-each proc ts)))
)
