;;; $Id: tc-tracing.scm,v 1.4 2006/01/18 23:17:27 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Tracing level for the type checker
;;; AUTHORS: Gary T. Leavens and Brian Dorn

;;;----------------------------------------------------------------------------
;;;
;;; An environment is an object that associates variables to values.

(module tc-tracing (lib "typedscm.ss" "typedscm")

(provide type-tracing-set! type-tracing-off! type-tracing-level
	 tc:type-tracing-set!)

(define *tc:tracing-level* 0)

;;; The following set the level of tracing in the type checker
(deftype type-tracing-set! (-> (number) void))
(define type-tracing-set!
  (lambda (num)
    (tc:type-tracing-set! num)
    (if (> *tc:tracing-level* 0)
	(displayln "Typed interpreter verbose tracing is now on at level "
		   *tc:tracing-level*)
	(displayln "Typed interpreter verbose tracing is now off."))))

(deftype tc:type-tracing-set! (-> (number) void))
(define tc:type-tracing-set!
  (lambda (num)
    (set! *tc:tracing-level* num)))

;;; The following turns off tracing in the type checker
(deftype type-tracing-off! (-> () void))
(define type-tracing-off!
  (lambda ()
    (tc:type-tracing-set! 0)
    (displayln "Typed interpreter verbose tracing is now off.")))

;;; The following queries the status of the tracing level in the
;;; type checker.
(deftype type-tracing-level (-> () number))
(define type-tracing-level
  (lambda ()
    *tc:tracing-level*))

) ;; end module
